{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE MultiWayIf #-}
module Data.Find.Interpolation
 ( search, searchM
 ) where

import qualified Data.Find.Linear as Linear
import           Data.Functor.Identity
import           Data.Ratio

search :: forall ix a
        . (Ord ix, Integral ix, Ord a)
       => (ix -> a) -- ^ Read a value at
       -> (a -> Integer)
       -> (ix, ix) -- ^ (min, max)
       -> a -- ^ The value we're looking for
       -> Maybe ix
search r j rng t = runIdentity $ searchM (pure . r) j rng t

searchM :: forall m ix a
         . (Monad m, Ord ix, Integral ix, Ord a)
        => (ix -> m a) -- ^ Read a value at
        -> (a -> Integer)
        -> (ix, ix) -- ^ (min, max)
        -> a -- ^ The value we're looking for
        -> m (Maybe ix)
searchM r j' (rmin, rmax) t = do
    vl <- r rmin
    vh <- r rmax
    -- Setup our invariant.
    if | vl == t -> pure $ Just rmin
       | vh == t -> pure $ Just rmax
       | t < vl || vh < t -> pure $ Nothing
       | otherwise -> go (rmin, vl) (rmax, vh)
  where
    j :: a -> Rational
    j = fromIntegral . j'
    scanSize = 10
    go :: (ix, a) -> (ix, a) -> m (Maybe ix)
    go (l, _) (h, _) | l+scanSize >= h = Linear.searchM r (l, h) t
    go lc@(l, vl) hc@(h, vh) = do
      let m = max (l+1) $ min (h-1) $ floor $ ((toInteger l)%1) + (((j t) - (j vl)) * ((toInteger $ h-l)%1) / ((j vh) - (j vl)))
      mv <- r m
      case t `compare` mv of
        EQ -> pure $ Just m
        LT -> go lc      (m, mv)
        GT -> go (m, mv) hc
